import functools
import inspect
from abc import ABC
from typing import Callable, Generic

import websockets
from kilroy_ws_server_py_sdk.protocol import chat, handle_errors
from kilroy_ws_server_py_sdk.receivers import (
    NullReceiver,
    Receiver,
    ReturnType,
    SingleReceiver,
    StreamReceiver,
)
from kilroy_ws_server_py_sdk.senders import (
    DataType,
    Sender,
    SingleSender,
    StreamSender,
)
from kilroy_ws_server_py_sdk.utils import can_call_with_arg


class Handler(ABC, Generic[DataType, ReturnType]):
    _path: str
    _fn: Callable
    _receiver: Receiver[ReturnType]
    _sender: Sender[DataType]

    def __init__(
        self,
        path: str,
        receiver: Receiver[ReturnType],
        sender: Sender[DataType],
    ) -> None:
        super().__init__()
        self._path = path
        self._receiver = receiver
        self._sender = sender

    @property
    def path(self) -> str:
        return self._path

    @property
    def function(self) -> Callable:
        return self._fn

    async def handle(
        self,
        websocket: websockets.WebSocketServerProtocol,
        *args,
        **kwargs,
    ) -> None:
        async with chat(websocket) as chat_id:
            async with handle_errors(websocket, chat_id):
                incoming_data = self._receiver.receive(websocket, chat_id)
                try:
                    if can_call_with_arg(self._fn, "data"):
                        kwargs = {**kwargs, "data": incoming_data}
                    outgoing_data = self._fn(*args, **kwargs)
                    if inspect.isawaitable(outgoing_data):
                        outgoing_data = await outgoing_data
                    await self._sender.send(websocket, chat_id, outgoing_data)
                finally:
                    if inspect.iscoroutine(incoming_data):
                        incoming_data.close()

    def __call__(self, fn: Callable) -> "HandlerWrapper":
        self._fn = fn
        return HandlerWrapper(self)


class HandlerWrapper:
    def __init__(self, handler: Handler) -> None:
        self.handler = handler
        self._fn = handler.function

    def __get__(self, obj, type=None) -> "HandlerWrapper":
        self._fn = functools.partial(self._fn, obj)
        return self

    def __call__(self, *args, **kwargs):
        return self._fn(*args, **kwargs)


class Get(Handler):
    def __init__(self, path: str) -> None:
        super().__init__(path, receiver=NullReceiver(), sender=SingleSender())


class Subscribe(Handler):
    def __init__(self, path: str) -> None:
        super().__init__(path, receiver=NullReceiver(), sender=StreamSender())


class Request(Handler):
    def __init__(self, path: str) -> None:
        super().__init__(
            path, receiver=SingleReceiver(), sender=SingleSender()
        )


class RequestStreamIn(Handler):
    def __init__(self, path: str) -> None:
        super().__init__(
            path, receiver=StreamReceiver(), sender=SingleSender()
        )


class RequestStreamOut(Handler):
    def __init__(self, path: str) -> None:
        super().__init__(
            path, receiver=SingleReceiver(), sender=StreamSender()
        )


class RequestStreamInOut(Handler):
    def __init__(self, path: str) -> None:
        super().__init__(
            path, receiver=StreamReceiver(), sender=StreamSender()
        )
