from contextlib import asynccontextmanager
from functools import partial
from json import JSONDecodeError
from typing import AsyncIterable, Type, TypeVar
from uuid import UUID

import websockets
from pydantic import ValidationError

from kilroy_ws_server_py_sdk.errors import (
    AppError,
    CONVERSATION_ERROR,
    INTERNAL_SERVER_ERROR,
    INVALID_MESSAGE_ERROR,
    ProtocolError,
)
from kilroy_ws_server_py_sdk.messages import (
    AppErrorMessage,
    BaseMessage,
    DataMessage,
    ProtocolErrorMessage,
    StartMessage,
    StopMessage,
    StreamEndMessage,
)
from kilroy_ws_server_py_sdk.types import JSON

MessageType = TypeVar("MessageType", bound=BaseMessage)


def parse_message(data: str, message_type: Type[MessageType]) -> MessageType:
    try:
        return message_type.parse_raw(data)
    except ValidationError as e:
        raise INVALID_MESSAGE_ERROR from e


parse_start_message = partial(parse_message, message_type=StartMessage)
parse_data_message = partial(parse_message, message_type=DataMessage)
parse_stream_end_message = partial(
    parse_message, message_type=StreamEndMessage
)


def create_data_message(chat_id: UUID, payload: JSON) -> DataMessage:
    try:
        return DataMessage(chat_id=chat_id, payload=payload)
    except ValidationError as e:
        raise ProtocolError("Can't create data message.") from e


def serialize_data_message(message: DataMessage) -> str:
    try:
        return message.json()
    except JSONDecodeError as e:
        raise ProtocolError("Can't serialize data message.") from e


async def receive_start_message(
    websocket: websockets.WebSocketServerProtocol,
) -> StartMessage:
    data = await websocket.recv()
    return parse_start_message(data)


async def send_stop_message(
    websocket: websockets.WebSocketServerProtocol, chat_id: UUID
) -> None:
    await websocket.send(StopMessage(chat_id=chat_id).json())


def check_chat_id(message: BaseMessage, expected_id: UUID) -> None:
    if message.chat_id != expected_id:
        raise CONVERSATION_ERROR


async def receive_data_message(
    websocket: websockets.WebSocketServerProtocol, chat_id: UUID
) -> DataMessage:
    data = await websocket.recv()
    message = parse_data_message(data)
    check_chat_id(message, chat_id)
    return message


async def receive_data_message_stream(
    websocket: websockets.WebSocketServerProtocol, chat_id: UUID
) -> AsyncIterable[DataMessage]:
    async for data in websocket:
        try:
            message = parse_data_message(data)
        except ProtocolError as e:
            try:
                message = parse_stream_end_message(data)
            except ProtocolError:
                raise e

            check_chat_id(message, chat_id)
            return

        check_chat_id(message, chat_id)
        yield message


async def send_data_message(
    websocket: websockets.WebSocketServerProtocol,
    chat_id: UUID,
    payload: JSON,
) -> None:
    message = create_data_message(chat_id, payload)
    data = serialize_data_message(message)
    await websocket.send(data)


async def send_data_message_stream(
    websocket: websockets.WebSocketServerProtocol,
    chat_id: UUID,
    payloads: AsyncIterable[JSON],
) -> None:
    async for payload in payloads:
        message = create_data_message(chat_id, payload)
        data = serialize_data_message(message)
        await websocket.send(data)

    stream_end_message = StreamEndMessage(chat_id=chat_id)
    await websocket.send(stream_end_message.json())


@asynccontextmanager
async def chat(websocket: websockets.WebSocketServerProtocol) -> UUID:
    try:
        start = await receive_start_message(websocket)
    except ProtocolError as e:
        await websocket.send(ProtocolErrorMessage(reason=str(e)))
        return

    yield start.chat_id

    await send_stop_message(websocket, start.chat_id)


@asynccontextmanager
async def handle_errors(
    websocket: websockets.WebSocketServerProtocol, chat_id: UUID
) -> None:
    internal_server_error_message = AppErrorMessage(
        chat_id=chat_id,
        code=INTERNAL_SERVER_ERROR.code,
        reason=INTERNAL_SERVER_ERROR.reason,
    )

    to_raise = None

    try:
        yield
        return
    except ProtocolError as e:
        error_message = ProtocolErrorMessage(chat_id=chat_id, reason=str(e))
    except AppError as e:
        error_message = AppErrorMessage(
            chat_id=chat_id, code=e.code, reason=e.reason
        )
    except Exception as to_raise:
        error_message = internal_server_error_message

    try:
        data = error_message.json()
    except JSONDecodeError as to_raise:
        data = internal_server_error_message.json()

    await websocket.send(data)

    if to_raise is not None:
        raise to_raise
