from abc import ABC, abstractmethod
from typing import AsyncIterable, Awaitable, Generic, TypeVar
from uuid import UUID

import websockets

from kilroy_ws_server_py_sdk.protocol import (
    receive_data_message,
    receive_data_message_stream,
)
from kilroy_ws_server_py_sdk.types import JSON

ReturnType = TypeVar(
    "ReturnType", Awaitable[None], Awaitable[JSON], AsyncIterable[JSON]
)


class Receiver(ABC, Generic[ReturnType]):
    @abstractmethod
    def receive(
        self,
        websocket: websockets.WebSocketServerProtocol,
        chat_id: UUID,
    ) -> ReturnType:
        pass


class NullReceiver(Receiver[Awaitable[None]]):
    async def receive(
        self,
        websocket: websockets.WebSocketServerProtocol,
        chat_id: UUID,
    ) -> None:
        return None


class SingleReceiver(Receiver[Awaitable[JSON]]):
    async def receive(
        self,
        websocket: websockets.WebSocketServerProtocol,
        chat_id: UUID,
    ) -> JSON:
        message = await receive_data_message(websocket, chat_id)
        return message.payload


class StreamReceiver(Receiver[AsyncIterable[JSON]]):
    async def receive(
        self,
        websocket: websockets.WebSocketServerProtocol,
        chat_id: UUID,
    ) -> AsyncIterable[JSON]:
        async for message in receive_data_message_stream(websocket, chat_id):
            yield message.payload
