import asyncio
import logging
from http import HTTPStatus
from typing import Dict, List, Optional, Tuple

import websockets
from websockets.datastructures import Headers, HeadersLike
from websockets.legacy.server import WebSocketServerProtocol
from yrouter import NoMatch, Router, RouterConfigurationError, route
from yrouter.match import FullMatch
from yrouter.route_node import RouteNode

from kilroy_ws_server_py_sdk.controller import Controller
from kilroy_ws_server_py_sdk.handlers import Handler
from kilroy_ws_server_py_sdk.utils import surround


class Server:
    def __init__(
        self,
        controller: Controller,
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> None:
        self._controller = controller
        self._router = Router(
            self._build_route_tree(controller.get_handlers())
        )
        self._logger = logger

    @staticmethod
    def _group_by_prefix(
        handlers: List[Handler], prefix: str
    ) -> Dict[str, List[Handler]]:
        groups = {}
        for handler in handlers:
            group = surround(handler.path).removeprefix(prefix).split("/")[0]
            if group not in groups:
                groups[group] = []
            groups[group].append(handler)
        return groups

    @staticmethod
    def _find_root_handler(nodes: List[RouteNode]) -> Optional[RouteNode]:
        return next(
            (node.handler for node in nodes if node.component in {"", "/"}),
            None,
        )

    def _build_route_tree(
        self, handlers: List[Handler], prefix: str = "/"
    ) -> List[RouteNode]:
        prefix = surround(prefix)

        if len(handlers) == 0:
            return []
        if len(handlers) == 1:
            path = surround(handlers[0].path).removeprefix(prefix)
            return [route(path, handlers[0].handle)]

        groups = self._group_by_prefix(handlers, prefix)

        if len(groups) == 1:
            raise RouterConfigurationError("Some routes are duplicated.")

        routes = []

        for group, handlers in groups.items():
            if len(handlers) <= 1:
                routes.extend(self._build_route_tree(handlers, prefix))
            else:
                sub = self._build_route_tree(handlers, prefix + group + "/")
                handler = self._find_root_handler(sub)
                routes.append(route(group, handler=handler, subroutes=sub))

        return routes

    async def process(
        self, path: str, headers: Headers
    ) -> Optional[Tuple[HTTPStatus, HeadersLike, bytes]]:
        if self._router.match(path) is NoMatch:
            return HTTPStatus.NOT_FOUND, [], b"Path not found\n"

    async def handle(self, websocket: WebSocketServerProtocol) -> None:
        try:
            match = self._router.match(websocket.path)
            if not isinstance(match, FullMatch):
                return
            await match.handler(websocket, self._controller, **match.kwargs)
        except Exception as e:
            self._logger.warning(
                f"Exception occurred during handling of {websocket.path}",
                exc_info=e,
            )

    async def run(self, *args, **kwargs) -> None:
        async with websockets.serve(
            self.handle, *args, process_request=self.process, **kwargs
        ):
            await asyncio.Future()
